
import math
import numpy as np
import thread
import threading
import random
import sys
import wx
import wx.richtext as rt
from wx.lib.pubsub import pub

from wx import glcanvas
import OpenGL
from OpenGL.GL import *
from OpenGL.GL.ARB.shader_objects import *
from OpenGL.GL.ARB.vertex_shader import *
from OpenGL.GL.ARB.vertex_array_object import *
from OpenGL.GL.ARB.draw_instanced import *
from OpenGL.GL.ARB.instanced_arrays import *
from OpenGL.GL import shaders

import MatrixUtil as matrix

from PyTAS3DAssembly import *

from ctypes import sizeof, c_float, c_void_p, c_uint

RAND_SEED = -1

GRID_WIDTH  = 100
GRID_HEIGHT = 100
GRID_DEPTH  = 100
GROWTH_INC  = 400
CELLSIZE  = 60 #96
IMG_OFFSET_HORIZ = 0#17
IMG_OFFSET_VERT  = 0#17

PYTAS_FULLDISPLAY      = 0
PYTAS_NODISPLAY        = 1

EMPTY_LOC           = 0
FRONTIER_LOC        = 1
FRONTIER_LOC_OFF    = 2

#assertMode = wx.PYAPP_ASSERT_DIALOG

# IDs
PYTAS_ID_STOP          = 100
PYTAS_ID_FOR           = 101
PYTAS_ID_FF            = 102
PYTAS_ID_ZOOMIN        = 103
PYTAS_ID_ZOOMOUT       = 104
PYTAS_ID_ZOOM_MODE     = 105
PYTAS_ID_SELECT_MODE   = 106
PYTAS_ID_DELETE        = 107
PYTAS_ID_UNDO          = 108
PYTAS_ID_STEPCOUNT     = 109

PYTAS_MODE_ZOOM   = 0
PYTAS_MODE_SELECT = 1

PYTAS_SM_STEP = 0
PYTAS_SM_FULL = 1

CUBE_OFFS = [
    (1,1,1),(0,1,1),(0,0,1),(1,0,1),
    (0,0,0),(0,1,0),(1,1,0),(1,0,0),
    (1,1,1),(1,1,0),(0,1,0),(0,1,1),
    (0,0,0),(1,0,0),(1,0,1),(0,0,1),
    (1,1,1),(1,0,1),(1,0,0),(1,1,0),
    (0,0,0),(0,0,1),(0,1,1),(0,1,0),
]

# vertices to be used by OpenGL
# first three components are position
# second three components are normal
CUBE_VERTICES = np.array([
    # top
    0, 0, 1, 0, 0, 1,
    0, 1, 1, 0, 0, 1,
    1, 0, 1, 0, 0, 1,
    1, 1, 1, 0, 0, 1,
    
    # bottom
    0, 0, 0, 0, 0,-1,
    0, 1, 0, 0, 0,-1,
    1, 0, 0, 0, 0,-1,
    1, 1, 0, 0, 0,-1,

    # left
    0, 0, 0,-1, 0, 0,
    0, 0, 1,-1, 0, 0,
    0, 1, 0,-1, 0, 0,
    0, 1, 1,-1, 0, 0,

    # right
    1, 0, 0, 1, 0, 0,
    1, 0, 1, 1, 0, 0,
    1, 1, 0, 1, 0, 0,
    1, 1, 1, 1, 0, 0,

    # front
    0, 0, 0, 0,-1, 0,
    0, 0, 1, 0,-1, 0,
    1, 0, 0, 0,-1, 0,
    1, 0, 1, 0,-1, 0,

    # back
    0, 1, 0, 0, 1, 0,
    0, 1, 1, 0, 1, 0,
    1, 1, 0, 0, 1, 0,
    1, 1, 1, 0, 1, 0,
], dtype=np.float32)

CUBE_INDICES = np.array([
    # top
    0, 2, 1, 3, 1, 2,
    # bottom
    4, 5, 6, 7, 6, 5,
    # left
    8, 9, 10, 11, 10, 9,
    # right
    12, 14, 13, 15, 13, 14,
    # front
    16, 18, 17, 19, 17, 18,
    # back
    20, 21, 22, 23, 22, 21,
], dtype=np.uint32)

OUTLINE_VERTICES = np.array([
    0, 0, 0, 1, 0, 0,
    0, 0, 0, 0, 1, 0,
    0, 0, 0, 0, 0, 1,
    1, 1, 1, 1, 1, 0,
    1, 1, 1, 1, 0, 1,
    1, 1, 1, 0, 1, 1,
    0, 0, 1, 0, 1, 1,
    0, 0, 1, 1, 0, 1,
    1, 1, 0, 0, 1, 0,
    1, 1, 0, 1, 0, 0,
    0, 1, 0, 0, 1, 1,
    1, 0, 0, 1, 0, 1,
], dtype=np.float32)

COMPASS_VERTICES = np.array([
    0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0,
    0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0,
    0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1,
], dtype=np.float32)

# constants for OpenGL buffers
INT_SIZE   = sizeof(c_int)
UINT_SIZE  = sizeof(c_uint)
FLOAT_SIZE = sizeof(c_float)

POSITION_OFFSET = c_void_p(0 * FLOAT_SIZE)
NORMAL_OFFSET   = c_void_p(3 * FLOAT_SIZE)
OFFSET_OFFSET   = c_void_p(0 * INT_SIZE)
COLOR_OFFSET    = c_void_p(3 * INT_SIZE)
VERTEX_SIZE     = 3 * FLOAT_SIZE

# OpenGL shader attribute locations
POSITION_LOC = 0
NORMAL_LOC   = 1
OFFSET_LOC   = 2
COLOR_LOC    = 3

gl_major_version = 0
gl_minor_version = 0
gl_33 = False

class Compass:

    def __init__(self):
        pass

    def initGraphics(self):
        self.vertexArray = glGenVertexArrays(1)
        self.vertexBuffer = glGenBuffers(1)

        glBindVertexArray(self.vertexArray)

        glBindBuffer(GL_ARRAY_BUFFER, self.vertexBuffer)
        glBufferData(GL_ARRAY_BUFFER, len(COMPASS_VERTICES) * FLOAT_SIZE, COMPASS_VERTICES, GL_STATIC_DRAW)

        offset = c_void_p(0 * FLOAT_SIZE)
        glEnableVertexAttribArray(POSITION_LOC)
        glVertexAttribPointer(POSITION_LOC, 3, GL_FLOAT, GL_FALSE, 2 * VERTEX_SIZE, offset)
        offset = c_void_p(3 * FLOAT_SIZE)
        glEnableVertexAttribArray(COLOR_LOC)
        glVertexAttribPointer(COLOR_LOC, 3, GL_FLOAT, GL_FALSE, 2 * VERTEX_SIZE, offset)

        glBindVertexArray(0)

    def Draw(self):
        glLineWidth(2.0)
        glBindVertexArray(self.vertexArray)
        glDrawArrays(GL_LINES, 0, 6)
        glBindVertexArray(0)

class Partition:
    # A partition will be 32 x 32 x 32 tiles
    SIZE = 32
    DIMENSIONS = [SIZE] * 3
    NUM_TILES = SIZE**3

    TILE_DATA_SIZE = 3 * INT_SIZE + 3 * INT_SIZE # 3 ints for color, 3 ints for offset

    def __init__(self, offset):
        self.offset = [x * Partition.SIZE for x in offset]

        self.modified = False
        self.tiles = {}

    def initGraphics(self):
        self.vertexArray    = glGenVertexArrays(1)
        self.cubeBuffer     = glGenBuffers(1) # Vertex buffer to contain cube vertex info
        self.indexBuffer    = glGenBuffers(1) # Vertex buffer to contain cube index info
        self.instanceBuffer = glGenBuffers(1) # Vertex buffer to contain instancing info

        self.lineArray      = glGenVertexArrays(1)
        self.lineBuffer     = glGenBuffers(1)

        glBindVertexArray(self.vertexArray)

        glBindBuffer(GL_ARRAY_BUFFER, self.cubeBuffer)
        glBufferData(GL_ARRAY_BUFFER, len(CUBE_VERTICES) * FLOAT_SIZE, CUBE_VERTICES, GL_STATIC_DRAW)

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self.indexBuffer)
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, len(CUBE_INDICES) * UINT_SIZE, CUBE_INDICES, GL_STATIC_DRAW)

        glEnableVertexAttribArray(POSITION_LOC)
        glVertexAttribPointer(POSITION_LOC, 3, GL_FLOAT, GL_FALSE, 2 * VERTEX_SIZE, POSITION_OFFSET)
        glEnableVertexAttribArray(NORMAL_LOC)
        glVertexAttribPointer(NORMAL_LOC, 3, GL_FLOAT, GL_FALSE, 2 * VERTEX_SIZE, NORMAL_OFFSET)

        data = np.zeros(Partition.TILE_DATA_SIZE * Partition.NUM_TILES, dtype=np.int32)

        glBindBuffer(GL_ARRAY_BUFFER, self.instanceBuffer)
        glBufferData(GL_ARRAY_BUFFER, Partition.NUM_TILES * Partition.TILE_DATA_SIZE, data, GL_DYNAMIC_DRAW)

        glEnableVertexAttribArray(OFFSET_LOC)
        glVertexAttribIPointer(OFFSET_LOC, 3, GL_INT, Partition.TILE_DATA_SIZE, OFFSET_OFFSET)
        
        glEnableVertexAttribArray(COLOR_LOC)
        glVertexAttribIPointer(COLOR_LOC, 3, GL_INT, Partition.TILE_DATA_SIZE, COLOR_OFFSET)

        if gl_33:
            glVertexAttribDivisor(POSITION_LOC, 0)
            glVertexAttribDivisor(NORMAL_LOC, 0)
            glVertexAttribDivisor(OFFSET_LOC, 1)
            glVertexAttribDivisor(COLOR_LOC, 1)
        else:
            glVertexAttribDivisorARB(POSITION_LOC, 0)
            glVertexAttribDivisorARB(NORMAL_LOC, 0)
            glVertexAttribDivisorARB(OFFSET_LOC, 1)
            glVertexAttribDivisorARB(COLOR_LOC, 1)

        # Array to render outline of cube
        glBindVertexArray(self.lineArray)

        glBindBuffer(GL_ARRAY_BUFFER, self.lineBuffer)
        glBufferData(GL_ARRAY_BUFFER, len(OUTLINE_VERTICES) * FLOAT_SIZE, OUTLINE_VERTICES, GL_STATIC_DRAW)

        glEnableVertexAttribArray(POSITION_LOC)
        glVertexAttribPointer(POSITION_LOC, 3, GL_FLOAT, GL_FALSE, VERTEX_SIZE, POSITION_OFFSET)

        glBindBuffer(GL_ARRAY_BUFFER, self.instanceBuffer)
        #glBufferData(GL_ARRAY_BUFFER, Partition.NUM_TILES * Partition.TILE_DATA_SIZE, data, GL_DYNAMIC_DRAW)

        glEnableVertexAttribArray(OFFSET_LOC)
        glVertexAttribIPointer(OFFSET_LOC, 3, GL_INT, Partition.TILE_DATA_SIZE, OFFSET_OFFSET)
        
        glEnableVertexAttribArray(COLOR_LOC)
        glVertexAttribIPointer(COLOR_LOC, 3, GL_INT, Partition.TILE_DATA_SIZE, COLOR_OFFSET)

        if gl_33:
            glVertexAttribDivisor(POSITION_LOC, 0)
            glVertexAttribDivisor(OFFSET_LOC, 1)
            glVertexAttribDivisor(COLOR_LOC, 1)
        else:
            glVertexAttribDivisorARB(POSITION_LOC, 0)
            glVertexAttribDivisorARB(OFFSET_LOC, 1)
            glVertexAttribDivisorARB(COLOR_LOC, 1)

        glBindVertexArray(0)

        del data

    def deletePartition(self):
        #TODO: delete OpenGL variables here
        del self.tiles

    def AddTile(self, tile):
        pos = tuple([x - y for x, y in zip(tile.coords, self.offset)])

        # valid = [x >= 0 and x < 32 for x in pos]

        # if not all(valid):
        #     print "Error: tile added to incorrect partition. Position = " + str(pos) + ", Offset = " + str(self.offset)
        #     return False

        self.tiles[pos] = tile
        self.modified = True
        return True

    def RemoveTile(self, position):
        pos = tuple([x - y for x, y in zip(position, self.offset)])

        # valid = [x >= 0 and x < 32 for x in pos]
        # if not all(valid):
        #     print "Error: tile added to incorrect partition. Position = " + str(pos) + ", Offset = " + str(self.offset)
        #     return False

        if pos in self.tiles:
            del self.tiles[pos]
            self.modified = True
        return True

    def UpdateBuffers(self):
        if self.modified:
            data = np.array([], dtype=np.int32)
            for tile in self.tiles.itervalues():
                offset = np.array(tile.coords, dtype=np.int32)
                color = np.array(tile.type.colorVector, dtype=np.int32)
                data = np.append(data, offset)
                data = np.append(data, color)
            
            glBindVertexArray(self.vertexArray)
            glBindBuffer(GL_ARRAY_BUFFER, self.instanceBuffer)
            glBufferSubData(GL_ARRAY_BUFFER, 0, len(self.tiles) * Partition.TILE_DATA_SIZE, data)
            glBindVertexArray(0)

            self.modified = False

    def DrawTiles(self):
        if self.modified:
            self.UpdateBuffers()

        glBindVertexArray(self.vertexArray)
        if gl_33:
           glDrawElementsInstanced(GL_TRIANGLES, 36, GL_UNSIGNED_INT, None, len(self.tiles))
        else:
           glDrawElementsInstancedARB(GL_TRIANGLES, 36, GL_UNSIGNED_INT, None, len(self.tiles))
        glBindVertexArray(0)

    def DrawOutlines(self):
        if self.modified:
            self.UpdateBuffers()

        glLineWidth(2.0)

        glBindVertexArray(self.lineArray)
        if gl_33:
            glDrawArraysInstanced(GL_LINES, 0, 24, len(self.tiles))
        else:
            glDrawArraysInstancedARB(GL_LINES, 0, 24, len(self.tiles))
        glBindVertexArray(0)


#----------------------------------------------------------------------------

class PyTAS3DCanvas(glcanvas.GLCanvas):
    def __init__(self, parent, idIn, frame, size, app):
        attribList = [glcanvas.WX_GL_RGBA,
                      glcanvas.WX_GL_DOUBLEBUFFER,
                      glcanvas.WX_GL_DEPTH_SIZE, 24]
        glcanvas.GLCanvas.__init__(self, parent, idIn, size=size, style=wx.SUNKEN_BORDER, attribList=attribList)

        self.frame              = frame
        self.app                = app
        self.context            = glcanvas.GLContext(self)
        
        self.size               = self.GetClientSize()
        self.aspectRatio        = 1.0
        self.backgroundColor    = wx.Colour((255, 255, 128, 255))

        # Mouse state variables
        self.mousePosition      = None
        self.leftMouse          = False
        self.rightMouse         = False
        
        # Assembly state variables
        self.assembly           = PyTAS3DAssembly()
        self.keepGoing          = False
        self.stepMode           = PYTAS_SM_STEP
        self.displayPatterns    = []
        self.partitions         = {}
        self.newPartitionList   = []
        self.breakpoints        = []
        
        # Camera and orientation matricies for view
        self.projectionMatrix = matrix.identity_matrix()
        self.invProjMatrix    = matrix.identity_matrix()
        self.rotationMatrix   = matrix.identity_matrix()
        self.viewMatrix       = matrix.identity_matrix()
        self.cameraCenter     = np.array([0.0, 0.0, 0.0])
        self.cameraUp         = np.array([0.0, 1.0, 0.0])
        self.cameraPosition   = np.array([0.0, 0.0, 4.0])

        # OpenGL shader variables and attribute locations
        self.cubeShader    = None
        self.outlineShader = None
        self.compassShader = None
        self.screenShader  = None
        self.POSITION_LOC  = None
        self.NORMAL_LOC    = None
        self.OFFSET_LOC    = None
        self.COLOR_LOC     = None
        
        self.initialized = False

        self.compass = Compass()
        self.shiftDown = False

        self._lock = threading.Lock()
        self.SetRandomSeed(RAND_SEED)
        
        self.SubscribeEventHandlers()
        
        self.numTilesPerStep    = 1
        self.lastNumTilesUpdate = 0
        self.drawUpdates        = PYTAS_FULLDISPLAY
        self.savedUpdates       = []

        #self.Bind(wx.EVT_ERASE_BACKGROUND, lambda x:None)
        self.Bind(wx.EVT_MOUSE_EVENTS, self.OnMouse)
        self.Bind(wx.EVT_SIZE,         self.OnSize)
        self.Bind(wx.EVT_MOVE,         self.OnMove)
        self.Bind(wx.EVT_PAINT,        self.OnPaint)
        self.Bind(wx.EVT_CHAR,         self.OnChar)
        self.Bind(wx.EVT_KEY_DOWN,     self.OnKeyDown)
        self.Bind(wx.EVT_KEY_UP,       self.OnKeyUp)
        
    def SubscribeEventHandlers(self):
        pub.getDefaultPublisher().subscribe(self.UpdateStatusbar, "updateStatusbar")

    def UnsubscribeEventHandlers(self):
        pub.getDefaultPublisher().unsubscribe(self.UpdateStatusbar, "updateStatusbar")

    def SendStatusBarUpdate(self, field, msg):
        wx.CallAfter(pub.getDefaultPublisher().sendMessage, "updateStatusbar", msg = (field, msg))

    def UpdateStatusbar(self, msg):
        self.frame.SetStatusText(msg[1], msg[0])

    def SetStepMode(self, mode):
        self.stepMode = mode
        if mode == PYTAS_SM_FULL:
            self.app.SetNumTilesPerStepEnabled(False)
        else:
            self.app.SetNumTilesPerStepEnabled(True)

    def initShader(self, src):
        #if gl_33:
        #    vs_src_file = open(src + '.vert')
        #else:
        #    vs_src_file = open(src + '_compat.vert')

        vs_src_file = open(src + '.vert')
        vs_src = vs_src_file.read()
        vs_src_file.close()
        vertexShader = shaders.compileShader(vs_src, GL_VERTEX_SHADER)
        
        #if gl_33:
        #    fs_src_file = open(src + '.frag')
        #else:
        #    fs_src_file = open(src + '_compat.frag')

        fs_src_file = open(src + '.frag')
        fs_src = fs_src_file.read()
        fs_src_file.close()
        fragmentShader = shaders.compileShader(fs_src, GL_FRAGMENT_SHADER)

        shader = glCreateProgram()
        glAttachShader(shader, vertexShader)
        glAttachShader(shader, fragmentShader)

        glBindAttribLocation(shader, 0, "a_Position")
        glBindAttribLocation(shader, 1, "a_Normal")
        glBindAttribLocation(shader, 2, "a_TileOffset")
        glBindAttribLocation(shader, 3, "a_Color")

        glDeleteShader(vertexShader)
        glDeleteShader(fragmentShader)

        glLinkProgram(shader)

        return shader

    def initGraphics(self):
        self.initialized = True
        
        self.CameraLookAt()
        
        # GL initialization

        gl_major_version = glGetIntegerv(GL_MAJOR_VERSION)
        gl_minor_version = glGetIntegerv(GL_MINOR_VERSION)

        gl_33 = (gl_major_version > 3) or (gl_major_version == 3 and gl_minor_version >= 3)

        print "GL VERSION: " + str(gl_major_version) + "." + str(gl_minor_version)

        self.DoSetViewport()

        # Shaders

        self.cubeShader    = self.initShader("./shaders/pass")
        self.outlineShader = self.initShader("./shaders/outline")
        self.compassShader = self.initShader("./shaders/compass")

        self.compass.initGraphics()

        glEnable(GL_CULL_FACE)
        glClearColor(0.3, 0.3, 0.4, 1.0)

    def Reset(self, bFromFiles=True, bForce=False):
        if not bForce == True:
            dlg = wx.TextEntryDialog(self, "Are you sure you want to reset your system\nand abandon the current assembly?\nIf so, you can set your random seed below\n(use -1 to choose a new random value):", caption="Set random seed", value=str(self.randSeed))
            if not dlg.ShowModal() == wx.ID_OK:
                return
            result = int(dlg.GetValue())
            dlg.Destroy()
            self.SetRandomSeed(result)
    
            self.app.OutputMessage("Resetting tile assembly system...")
        
        if bFromFiles == True:
            defFile = None
            if (not self.assembly.tileSystemDef is None) and (not self.assembly.tileSystemDef['MainFile'] is None):
                defFile = self.assembly.tileSystemDef['MainFile']
            self.assembly = PyTAS3DAssembly()
            if not defFile is None:
                self.assembly.LoadTileSystem(defFile)
        else:
            self.assembly.ClearSpace()
            self.assembly.PlaceSeed()
        
        for partition in self.partitions.itervalues():
            partition.deletePartition()
        del self.partitions
        self.partitions = {}

        self.RedrawAssembly()
        cacheHitPercent = 0
        if self.assembly.cacheSearches > 0:
            cacheHitPercent = float(self.assembly.cacheHits/self.assembly.cacheSearches) * 100
        self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles))
        self.CameraLookAt()
        self.app.OutputMessage("Tile assembly system reset")

    def DoGo(self):
        self.app.SetNumTilesPerStepEnabled(False)
        if self.keepGoing:
            PrintDebug("ERROR: Cannot create another thread, there is already one running!", ERROR_MSG)
            return
        self.keepGoing = True
        try:
            self.goThread = thread.start_new_thread(self.GoForward, ())
        except:
            PrintDebug("ERROR: Failed to create thread!", ERROR_MSG)

    def GoForward(self):
        while self.keepGoing:
            if not self.DoForwardStep():
                self.keepGoing = False
                cacheHitPercent = 0
                if self.assembly.cacheSearches > 0:
                    cacheHitPercent = float(self.assembly.cacheHits/self.assembly.cacheSearches) * 100
                self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles))
        self.app.SetNumTilesPerStepEnabled(True)
        PrintDebug("Stopping the thread!", MSG_MSG)

    def DoForwardStep(self):
        if self.stepMode == PYTAS_SM_STEP:
            steps = self.numTilesPerStep
            if steps < 1:
                steps = 1
            for i in range(steps):
                newLoc = self.assembly.StepForward()
                if not newLoc is None:
                    if self.drawUpdates == PYTAS_FULLDISPLAY:
                        self.AddToPartition(newLoc[0])
                    elif self.drawUpdates == PYTAS_NODISPLAY:
                        self._lock.acquire()
                        self.savedUpdates.append(newLoc[0])
                        self._lock.release()
                    if (self.keepGoing == False) or ((self.assembly.numTiles - self.lastNumTilesUpdate) >= 1000):
                        self.lastNumTilesUpdate = self.assembly.numTiles
                    #if ((self.assembly.numTiles % 1000) >= 0) and ((self.assembly.numTiles % 1000) < steps):

                    #elif ( ((self.assembly.numTiles - lastNumTiles) % 1000) == 0):
                        #lastNumTiles = self.assembly.numTiles
                        cacheHitPercent = 0
                        if self.assembly.cacheSearches > 0:
                            cacheHitPercent = float(self.assembly.cacheHits/self.assembly.cacheSearches) * 100
                        self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles))
                    if newLoc[1] == False:
                        self.Refresh()
                        return False
                    self.CheckForBreakpointHit(newLoc[0])
                else:
                    self.Refresh()
                    return False
            if self.keepGoing == False:
                cacheHitPercent = 0
                if self.assembly.cacheSearches > 0:
                    cacheHitPercent = float(self.assembly.cacheHits/self.assembly.cacheSearches) * 100
                self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles))
            if self.drawUpdates == PYTAS_FULLDISPLAY:
                self.Refresh()
            return True
        elif self.stepMode == PYTAS_SM_FULL:
            newLocs = self.assembly.StepForward(True)
            if (self.keepGoing == False) or ((self.assembly.numTiles - self.lastNumTilesUpdate) >= 1000):
                self.lastNumTilesUpdate = self.assembly.numTiles
                cacheHitPercent = 0
                if self.assembly.cacheSearches > 0:
                    cacheHitPercent = float(self.assembly.cacheHits/self.assembly.cacheSearches) * 100
                self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles))
            if not newLocs is None:
                for newLoc in newLocs[0]:
                    if self.drawUpdates == PYTAS_FULLDISPLAY:
                        self.AddToPartition(newLoc)
                    elif self.drawUpdates == PYTAS_NODISPLAY:
                        self._lock.acquire()
                        self.savedUpdates.append(newLoc)
                        self._lock.release()
                self.CheckForBreakpointHit(newLoc)
                if self.keepGoing == False:
                    cacheHitPercent = 0
                    if self.assembly.cacheSearches > 0:
                        cacheHitPercent = float(self.assembly.cacheHits/self.assembly.cacheSearches) * 100
                    self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles))
                self.Refresh()
                return newLocs[1]
            return False

    def CheckForBreakpointHit(self, newTile):
        if len(self.breakpoints) > 0:
            for bp in self.breakpoints:
                if bp.type == SimUtils.SIMU_BP_TYPE_TILENUM:
                    if self.assembly.numTiles == bp.tileNum:
                        self.keepGoing = False
                        SimUtils.globalApp.OutputMessage("Hit breakpoint for " + str(self.assembly.numTiles) + " tiles in the assembly.")
                elif bp.type == SimUtils.SIMU_BP_TYPE_TILETYPE:
                    if bp.tileType == newTile[1]:
                        self.keepGoing = False
                        SimUtils.globalApp.OutputMessage("Hit breakpoint for tile of type " + bp.tileType + " placed in the assembly.")
                else:
                    newCoords = newTile[0]
                    bHit = False
                    bKeepChecking = True
                    if (bp.tileLoc[0] is not None):
                        if (bp.tileLoc[0] == newCoords[0]):
                            bHit = True
                        else:
                            bKeepChecking = False
                    if (bKeepChecking == True) and (bp.tileLoc[1] is not None):
                        if (bp.tileLoc[1] == newCoords[1]):
                            bHit = True
                        else:
                            bHit = False
                            bKeepChecking = False
                    if (bKeepChecking == True) and (bp.tileLoc[2] is not None):
                        if (bp.tileLoc[2] == newCoords[2]):
                            bHit = True
                        else:
                            bHit = False
                    if bHit == True:
                        self.keepGoing = False
                        SimUtils.globalApp.OutputMessage("Hit breakpoint for tile placed in location " + str(newCoords) + " which matches " + str(bp.tileLoc))

    def CameraLookAt(self, eye = [0.0, 0.0, 8.0], center = [0.0, 0.0, 0.0], up = [0.0, 1.0, 0.0]):
        self.cameraPosition = np.array(eye)
        self.cameraCenter   = np.array(center)
        self.cameraUp       = np.array(up)

        self.viewMatrix = matrix.look_at_matrix(self.cameraPosition, self.cameraCenter, self.cameraUp)

    def OnPaint(self, event):
        if self.drawUpdates == PYTAS_FULLDISPLAY:
            self.savedUpdates.reverse()
            while len(self.savedUpdates) > 0:
                self.AddToPartition(self.savedUpdates.pop())

        self._lock.acquire()
        
        dc = wx.PaintDC(self)
        self.SetCurrent(self.context)
        
        if not self.initialized:
            self.initGraphics()

        if self.newPartitionList:
            for offset in self.newPartitionList:
                self.partitions[offset].initGraphics()
            self.newPartitionList = []

        glEnable(GL_DEPTH_TEST)
        glDepthFunc(GL_LEQUAL)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        
        for partition in self.partitions.itervalues():
            glUseProgram(self.cubeShader)

            projMatLoc = glGetUniformLocation(self.cubeShader, "u_ProjectionMatrix")
            viewMatLoc = glGetUniformLocation(self.cubeShader, "u_ViewMatrix")

            glUniformMatrix4fv(projMatLoc, 1, GL_FALSE, self.projectionMatrix)
            glUniformMatrix4fv(viewMatLoc, 1, GL_FALSE, self.viewMatrix)

            partition.DrawTiles()

            glUseProgram(self.outlineShader)

            projMatLoc = glGetUniformLocation(self.outlineShader, "u_ProjectionMatrix")
            viewMatLoc = glGetUniformLocation(self.outlineShader, "u_ViewMatrix")

            glUniformMatrix4fv(projMatLoc, 1, GL_FALSE, self.projectionMatrix)
            glUniformMatrix4fv(viewMatLoc, 1, GL_FALSE, self.viewMatrix)

            partition.DrawOutlines()

        glDisable(GL_DEPTH_TEST)

        glUseProgram(self.compassShader)

        screenLoc = glGetUniformLocation(self.compassShader, "u_ScreenSize")
        offsetLoc = glGetUniformLocation(self.compassShader, "u_Offset")
        rotationLoc = glGetUniformLocation(self.compassShader, "u_ViewMatrix")
        glUniform2i(screenLoc, self.size.width, self.size.height)
        glUniform3f(offsetLoc, -0.9, -0.85, 0.0)
        glUniformMatrix4fv(rotationLoc, 1, GL_FALSE, self.viewMatrix)
        self.compass.Draw()

        glUseProgram(0)

        self.SwapBuffers()
        self._lock.release()

    def AddToPartition(self, (coords, tileName)):
        if self.NameMatchesPatterns(tileName):
            self._lock.acquire()

            offset = tuple([x // Partition.SIZE for x in coords])

            if offset not in self.partitions:
                self.partitions[offset] = Partition(offset)
                self.newPartitionList.append(offset)
            self.partitions[offset].AddTile(self.assembly.GetLoc(coords))

            self._lock.release()

    def RedrawAssembly(self, checkAll=False):
        self._lock.acquire()
        self.drawnAssembly = []

        for tile in self.assembly.grid.itervalues():
            if (not tile == FRONTIER_LOC) and (not tile == FRONTIER_LOC_OFF) and self.NameMatchesPatterns(tile.type.name):
                offset = tuple([x // Partition.SIZE for x in tile.coords])
                if offset not in self.partitions:
                    self.partitions[offset] = Partition(offset)
                    self.partitions[offset].initGraphics()
                self.partitions[offset].AddTile(tile)

        self._lock.release()
        self.Refresh()

    def DoSetViewport(self):
        size = self.size = self.GetClientSize()
        self.SetCurrent(self.context)
        glViewport(0, 0, size.width, size.height)
        self.aspectRatio = size.width * 1.0 / size.height
        self.projectionMatrix = matrix.perspective_matrix(60.0, self.aspectRatio, 1.0, 1000.0)
        self.invProjMatrix = np.linalg.inv(self.projectionMatrix)

    def NameMatchesPatterns(self, name):
        if len(self.displayPatterns) == 0:
            return True
        else:
            for pattern in self.displayPatterns:
                if pattern in name:
                    return True
            return False

    def DoQuit(self):
        self.keepGoing = False

    def SetRandomSeed(self, seed):
        self.randSeed = seed
        if self.randSeed < 0:
            self.randSeed = random.randint(0, sys.maxint)
        self.app.OutputMessage("Seeding random number generator with " + str(self.randSeed))
        random.seed(self.randSeed)

    def ScrollToCoords(self, coords):
        #This is not (yet?) implemented
        return

#===============================================================================
#     def DoLoad(self):
#         openFileDialog = wx.FileDialog(self.app.window, "Open", "", "", 
#                                        "Tile system files (*.tdp)|*.tdp",
#                                        wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
#         openFileDialog.ShowModal()
#         fileName = openFileDialog.GetPath()
#         openFileDialog.Destroy()
#         if (len(fileName) > 0):
#             
#             dlg = wx.TextEntryDialog(self, "You can set your random seed below\n(use -1 to choose a new random value):", caption="Set random seed", value=str(self.randSeed))
#             if dlg.ShowModal() == wx.ID_OK:
#                 self.randSeed = int(dlg.GetValue())
#                 dlg.Destroy()
#             else:
#                 dlg.Destroy()
#                 return
# 
#             print("I'm here!")
#             self.SetRandomSeed(self.randSeed)
#             self.app.OutputMessage("Loading tile assembly system...")
#             
#             #MJP
#             self.assembly.ClearSpace()
#             self.assembly.seedPatch = []
#             print("placing seed: " + str(self.assembly.seedPatch))
#             self.assembly.PlaceSeed()
# 
#             for partition in self.partitions.itervalues():
#                 partition.deletePartition()
#             del self.partitions
#             self.partitions = {}
#             
#             self.RedrawAssembly()
#             #MJP
#             
# 
#             self.assembly = PyTAS3DAssembly()
#             self.assembly.LoadTileSystem(fileName, backgroundColor=self.backgroundColor)
# 
#             self.app.OutputMessage("Tile assembly system loaded")
# 
# 
#             self.assembly.ClearSpace()
#             self.assembly.PlaceSeed()
# 
#             for partition in self.partitions.itervalues():
#                 partition.deletePartition()
#             del self.partitions
#             self.partitions = {}
#             
#             self.RedrawAssembly()
#             self.Refresh()
#             self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles))
#===============================================================================

    def DoSave(self):
        openFileDialog = wx.FileDialog(self.app.window, "Save", "", "", "Tile system files (*.tdp)|*.tdp",
                                        wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        openFileDialog.ShowModal()
        fileName = openFileDialog.GetPath()
        openFileDialog.Destroy()
        if len(fileName) > 0:
            assemblyString = self.assembly.AssemblyToString()
            file = open(fileName, 'w')
            file.write(assemblyString)
            file.close()

    def OnSize(self, event):
        wx.CallAfter(self.DoSetViewport)
        wx.CallAfter(self.Refresh)

    def OnMove(self, event):
        wx.CallAfter(self.DoSetViewport)
        wx.CallAfter(self.Refresh)

    def SetBackgroundColor(self):
        colorDialog = wx.ColourDialog(self.frame)
        ok = colorDialog.ShowModal()
        color = colorDialog.GetColourData().GetColour()

        self.backgroundColor = color
        glClearColor(color.Red() / 255, color.Green() / 255, color.Blue() / 255, 1.0)

    def GetPartitionsOnScreen(self):
        # TODO calculate the partitions that are visible to the camera
        return self.partitions

    def GetMouseProjection(self, mousepos):
        x = (2.0 * mousepos[0]) / self.size.width - 1.0
        y = 1.0 - (2.0 * mousepos[1]) / self.size.height
        vec = np.array([x, y, -1.0, 1.0])

        # TO CAMERA COORDS
        vec = np.matmul(self.invProjMatrix, vec)
        vec[2] = -1.0
        vec[3] = 0.0

        # TO WORLD COORDS
        rotMatrix = np.array(self.viewMatrix)
        rotMatrix[3][0] = 0.0
        rotMatrix[3][1] = 0.0
        rotMatrix[3][2] = 0.0
        vec = np.matmul(rotMatrix, vec)
        vec = vec / np.linalg.norm(vec)

        dx = math.fabs(vec[0])
        dy = math.fabs(vec[1])
        dz = math.fabs(vec[2])

        x0 = self.cameraPosition[0]
        y0 = self.cameraPosition[1]
        z0 = self.cameraPosition[2]

        x1 = x0 + vec[0]
        y1 = y0 + vec[1]
        z1 = z0 + vec[2]

        x = int(math.floor(self.cameraPosition[0]))
        y = int(math.floor(self.cameraPosition[1]))
        z = int(math.floor(self.cameraPosition[2]))

        dt_dx = float('inf') if dx == 0 else 1.0 / dx
        dt_dy = float('inf') if dy == 0 else 1.0 / dy
        dt_dz = float('inf') if dz == 0 else 1.0 / dz

        t = 0
        n = 1

        if dx == 0:
            x_inc = 0
            t_next_x = dt_dx
        elif x1 > x0:
            x_inc = 1
            n += int(math.floor(x1)) - x
            t_next_x = (math.floor(x0) + 1 - x0) * dt_dx
        else:
            x_inc = -1
            n += x - int(math.floor(x1))

            if math.fabs(x0) > 0.00000001:
                t_next_x = (x0 - math.floor(x0)) * dt_dx
            else:
                t_next_x = 0.0 * dt_dx

        if dy == 0:
            y_inc = 0
            t_next_y = dt_dy
        elif y1 > y0:
            y_inc = 1
            n += int(math.floor(y1)) - y
            t_next_y = (math.floor(y0) + 1 - y0) * dt_dy
        else:
            y_inc = -1
            n += y - int(math.floor(y1))
            
            if math.fabs(y0) > 0.00000001:
                t_next_y = (y0 - math.floor(y0)) * dt_dy
            else:
                t_next_y = 0.0 * dt_dy

        if dz == 0:
            z_inc = 0
            t_next_z = dt_dz
        elif z1 > z0:
            z_inc = 1
            n += int(math.floor(z1)) - z
            t_next_z = (math.floor(z0) + 1 - z0) * dt_dz
        else:
            z_inc = -1
            n += z - int(math.floor(z1))
            
            if math.fabs(z0) > 0.00000001:
                t_next_z = (z0 - math.floor(z0)) * dt_dz
            else:
                t_next_z = 0.0 * dt_dz

        for i in range(200):

            loc = self.assembly.GetLoc((x,y,z))
            if loc is not None:
                return loc

            if t_next_x < t_next_y and t_next_x < t_next_z:
                x += x_inc
                t = t_next_x
                t_next_x += dt_dx
            elif t_next_y < t_next_x and t_next_y < t_next_z:
                y += y_inc
                t = t_next_y
                t_next_y += dt_dy
            elif t_next_z < t_next_x and t_next_z < t_next_y:
                z += z_inc
                t = t_next_z
                t_next_z += dt_dz

            n -= 1
        return None

    def OnChar(self, evt):
        keycode = evt.GetKeyCode()
        if keycode == ord("f") or keycode == ord("F"):
            self.DoForwardStep()
        #elif keycode == ord("p") or keycode == ord("P"):
            #self.PrintGrowth()
        elif keycode == ord("r") or keycode == ord("R"):
            self.Reset(False)
        elif keycode == ord("s") or keycode == ord("S"):
            self.DoSave()
        elif keycode == ord("h") or keycode == ord("H"):
            self.ShowHelp()
        elif keycode == ord("l") or keycode == ord("L"):
            self.DoLoad()
        elif keycode == ord("=") or keycode == ord("+"):
            self.ZoomThenRecenter(True)
        elif keycode == ord("-") or keycode == ord("_"):
            self.ZoomThenRecenter(False)
        elif keycode == ord("g") or keycode == ord("G"):
            self.DoGo()
        elif keycode == ord("q") or keycode == ord("Q"):
            self.DoQuit()
        elif keycode == ord("c") or keycode == ord("C"):
            self.ScrollToCoords((0,0))
        elif keycode == ord("t") or keycode == ord("T"):
            self.ToggleGrid()
        elif keycode == ord("d") or keycode == ord("D"):
            self.ToggleDisplay()
        elif keycode == wx.WXK_F2:
            self.Refresh()
        elif keycode == wx.WXK_RETURN:
            self.DeleteSelection()

        #elif keycode == wx.WXK_DELETE or keycode == wx.WXK_BACK:
            #self.DoDelete()

    def OnKeyDown(self, event):
        keycode = event.GetKeyCode()
        if keycode == wx.WXK_SHIFT:
            self.shiftDown = True

    def OnKeyUp(self, event):
        keycode = event.GetKeyCode()
        if keycode == wx.WXK_SHIFT:
            self.shiftDown = False

    def OnMouse(self, event):
        # coordinates of current mouse event
        newPosition = [event.GetX(), event.GetY()]
        scroll = event.GetWheelRotation()

        if self.assembly is not None:
            loc = self.GetMouseProjection(newPosition)
            if loc is not None:
                ttype = loc.type
                self.SendStatusBarUpdate(0, "Coords: " + str(loc.coords))
                msg = "Tile type: " + ttype.ToString(bFull=True)
#===============================================================================
#                 msg =  ttype.name
# 
#                 glues = []
#                 if ttype.sides[N].label is not "":
#                     glues.append(("N", ttype.sides[N].label, ttype.sides[N].strength))
#                 if ttype.sides[S].label is not "":
#                     glues.append(("S", ttype.sides[S].label, ttype.sides[S].strength))
#                 if ttype.sides[E].label is not "":
#                     glues.append(("E", ttype.sides[E].label, ttype.sides[E].strength))
#                 if ttype.sides[W].label is not "":
#                     glues.append(("W", ttype.sides[W].label, ttype.sides[W].strength))
#                 if ttype.sides[U].label is not "":
#                     glues.append(("U", ttype.sides[U].label, ttype.sides[U].strength))
#                 if ttype.sides[D].label is not "":
#                     glues.append(("D", ttype.sides[D].label, ttype.sides[D].strength))
# 
#                 if glues:
#                     msg += " ["
# 
#                 first = True
#                 for glue in glues:
#                     if first:
#                         first = False
#                     else:
#                         msg += " -- "
#                     msg += glue[0] + ": " + glue[1] + "(" + str(glue[2]) + ")"
#                 if glues:
#                     msg += "]"
#===============================================================================
            else:
                self.SendStatusBarUpdate(0, "")
                msg = ""
            self.SendStatusBarUpdate(2, msg)

        if event.LeftDown():
            self.leftMouse = True
        elif event.LeftUp():
            self.leftMouse = False
        elif event.RightDown():
            self.rightMouse = True
        elif event.RightUp():
            self.rightMouse = False
        elif event.Dragging():
            if self.leftMouse:
                if self.rightMouse or self.shiftDown:
                    radius = np.linalg.norm(self.cameraPosition)
                    movement = np.append(np.array(np.subtract(newPosition, self.mousePosition) * 4.0 / [self.size.width, -self.size.height]), 0.0)
                    offset = -matrix.sphere_translation(self.cameraPosition, self.cameraUp, movement)
                    self.cameraCenter += offset * radius * 0.5
                    self.viewMatrix = matrix.look_at_matrix(self.cameraPosition, self.cameraCenter, self.cameraUp)
                    self.Refresh()
                else:
                    M = matrix.sphere_rotation_matrix(self.cameraPosition, self.cameraUp, self.mousePosition, newPosition, self.size.width, self.size.height)
                    self.cameraPosition = np.matmul(M, self.cameraPosition)
                    self.cameraUp = matrix.normalize(np.matmul(M, self.cameraUp))
                    self.viewMatrix = matrix.look_at_matrix(self.cameraPosition, self.cameraCenter, self.cameraUp)
                    self.Refresh()
                
        elif not scroll == 0:
            if scroll < 0:
                self.cameraPosition /= 0.95
            else:
                self.cameraPosition /= 1.05
            self.viewMatrix = matrix.look_at_matrix(self.cameraPosition, self.cameraCenter, self.cameraUp)
            self.Refresh()
        # set coordinates of previous mouse event
        self.mousePosition = newPosition
